;*****************************************************************************
;* Naughty Tile-Based Background Renderer Version 0.0                        *
;* Public domain by Bart Trzynadlowski, 1999, 2000                           *
;*****************************************************************************

;
; nrender.asm
;
; This is a heavily modified, configurable version of the poor man's code that
; did the background rendering in the Naughty NES emulator. It is tile-based,
; not too accurate, but decent. It cannot handle scrolling without some major
; hacking to the usage of x, y, CURRENT_TILE_X, and CURRENT_TILE_Y.
;

; ** all the name definitions go here **
%define RENDER_BACKGROUND       _your_name_for_the_rendering_function
%define CTRL_REG_2              _your_name_for_the_ppu_control_register_2
%define BG_PTRN_TABLE           _your_name_for_the_background_pattern_table
%define NAME_TABLE              _your_name_for_the_name_table
%define IMAGE_PALETTE           _your_name_for_the_image_palette
%define ATTRIB_LOOKUP           _your_name_for_the_attribute_lookup_table
%define VIDEO_BUFFER            _your_name_for_the_video_double_buffer

; ** public data and code **
global                          RENDER_BACKGROUND
global                          VIDEO_BUFFER

; ** external dependencies **
extern                          CTRL_REG_2
extern                          BG_PTRN_TABLE
extern                          NAME_TABLE
extern                          IMAGE_PALETTE
extern                          ATTRIB_LOOKUP

; ** data **
align 8
section .data

        ; tiles to start rendering at (must always be 0)
align 8
        CURRENT_TILE_X  db 0
align 8
        CURRENT_TILE_Y  db 0

        ; video double buffer (keep 320 otherwise it crashes, render 224 or
        ; 240. without 320 it will crash)
align 8
        VIDEO_BUFFER    times (256 * 320) db 0

        ; various internal data
align 8
        x               dd 0
align 8
        y               dd 0
align 8
        bit_mask        db 0
align 8
        row             db 0      

; ** 32-bit code **
bits 32
section .text

; plot: Plots a pixel on the screen. Uses edx as a register for arithmetic.
%macro plot 3           ; %1 = x (32-bit) %2 = y (32-bit) %3 = 8-bit data
        push    edx
        mov     edx, %2
        shl     edx, 8
        add     edx, %1         ; y * 256 + x
        add     edx, VIDEO_BUFFER
        mov     [edx], %3
        pop     edx
%endmacro

;*****************************************************************************
; void RENDER_BACKGROUND():
; Input:        None.
; Returns:      Nothing.
; Notes:        This could use a lot of optimization, particularly the removal
;               of x and y variables and better usage of the attribute lookup
;               table. Excuse the poor comments.
;*****************************************************************************
align 8
RENDER_BACKGROUND:
        push    ebp
        mov     ebp, esp

        test    byte [CTRL_REG_2], 08h  ; is background visible?
        jz      near background_done    ; no, do not draw it

        xor     eax, eax
        xor     ebx, ebx

        ; set up x and y, * 8 because a tile is 8x8 
        mov     al, [CURRENT_TILE_X]
        mov     bl, [CURRENT_TILE_Y]
        shl     eax, 3
        shl     ebx, 3
        mov     [x], eax
        mov     [y], ebx

        ; get starting offset in name table
        shl     ebx, 2          ; << 3 << 2 = << 5 = * 32
        shr     eax, 3          ; get eax back to normal
        mov     esi, ebx
        add     esi, eax
        mov     eax, esi        ; save [cur_tile_y] * 32 + [cur_tile_x] here
        add     esi, [NAME_TABLE]      ; physical address

        ; get offset of proper entry in attribute look-up table
        mov     edi, eax
        shl     edi, 2
        add     edi, ATTRIB_LOOKUP        

        ; set up main counter
        mov     ecx, eax

main_loop:
        movzx   ebx, byte [esi] ; get index into pattern table
        shl     ebx, 4          ; * 16 (each pattern table entry is 16 bytes)
        add     ebx, [BG_PTRN_TABLE]

        push    ecx
        mov     ecx, 8          ; y loop count: 8, tiles are 8 pixels tall

        loop_y:
                mov     al, [ebx]       ; get first part of current tile line
                mov     dl, [ebx+8]     ; next part

                push    ebx
                push    ecx
                mov     byte [bit_mask], 80h    ; mask initialized for bit 7
                mov     ecx, 8                  ; x loop count

                loop_x:
                        mov     ah, al  ; don't modify original patterns
                        mov     dh, dl

                        and     ah, [bit_mask]  ; isolate current bit
                        and     dh, [bit_mask]

                        shr     ah, cl  ; shift right (too far, into CF)
                        setc    ah      ; get it out of CF
                        shr     dh, cl
                        setc    dh      ; same situation with dh
                        shl     dh, 1   ; shift left 1, it is bit 1 of color

                        or      ah, dh  ; combine them properly

                        ; get upper 2 color bits
                        xor     ebx, ebx
                        mov     bl, [edi]       ; attrib. table index
                        add     ebx, 3c0h       ; ntab + 3c0h = attrib. tab
                        add     ebx, [NAME_TABLE]
                        mov     dh, [ebx]       ; get attrib byte
                        push    ecx             ; save loop count
                        mov     cl, [edi+2]     ; get shr count
                        shr     dh, cl          ; isolate the proper bits
                        and     dh, 3
                        shl     dh, 2           ; align them properly
                        pop     ecx             ; restore loop counter

                        or      ah, dh          ; ah = palette index                       

                        xor     ebx, ebx
                        mov     bl, ah          ; color
                        add     ebx, [IMAGE_PALETTE]
                        mov     ah, [ebx]       ; get actual color

                        plot    [x], [y], ah    ; draw the pixel!

                        inc     dword [x]        ; next x position
                        shr     byte [bit_mask], 1      ; next bit
                dec     ecx
                cmp     ecx, 0
                jne     near loop_x             ; loop

                pop     ecx             ; restore y loop count
                pop     ebx             ; restore pattern index
                inc     ebx             ; next pattern line
                sub     dword [x], 8    ; back to proper x position
                inc     dword [y]       ; next y
        dec     ecx
        cmp     ecx, 0
        jne     near loop_y     ; loop
       
        pop     ecx             ; restore main loop count
        add     edi, 4          ; next attribute table index
        inc     esi             ; next name table thing
        add     dword [x], 8    ; next x 
        sub     dword [y], 8    ; reset y
        inc     byte [row]
        cmp     byte [row], 32  ; filled up a row of tiles?
        jne     continue_main   ; no, continue
                add     dword [y], 8    ; next y
                mov     dword [x], 0    ; start x over at beginning
                mov     byte [row], 0   ; start row over
continue_main:
        inc     ecx
        cmp     ecx, (32 * 30)          ; if we've hit last tile, stop
        jne     near main_loop

        mov     byte [CURRENT_TILE_X], 0        ; reset
        mov     byte [CURRENT_TILE_Y], 0
        mov     byte [row], 0
        
background_done:
        pop     ebp
ret
        

